"""
Provides a test case for data types required by issue 374 - "Support for Point property type".

At the moment, only one new datatype is offered: NeomodelPoint

For more information please see: https://github.com/neo4j-contrib/neomodel/issues/374
"""

import os
import neomodel
import neomodel.contrib.spatial_properties
import shapely
import pytest


def version_to_dec(a_version_string):
    """
    Converts a version string to a number to allow for quick checks on the versions of specific components.

    :param a_version_string: The version string under test (e.g. '3.4.0')
    :type a_version_string: str
    :return: An integer representation of the string version, e.g. '3.4.0' --> 340
    """
    components = a_version_string.split('.')
    while len(components) < 3:
        components.append('0')
    num = 0
    for a_component in enumerate(components):
        num += (10 ** ((len(components) - 1) - a_component[0])) * int(a_component[1])
    return num


def check_and_skip_neo4j_least_version(required_least_neo4j_version, message):
    """
    Checks if the NEO4J_VERSION is at least `required_least_neo4j_version` and skips a test if not.

    WARNING: If the NEO4J_VERSION variable is not set, this function returns True, allowing the test to go ahead.

    :param required_least_neo4j_version: The least version to check. This must be the numberic representation of the
    version. That is: '3.4.0' would be passed as 340.
    :type required_least_neo4j_version: int
    :param message: An informative message as to why the calling test had to be skipped.
    :type message: str
    :return: A boolean value of True if the version reported is at least `required_least_neo4j_version`
    """
    if 'NEO4J_VERSION' in os.environ:
        if version_to_dec(os.environ['NEO4J_VERSION']) < required_least_neo4j_version:
            pytest.skip('Neo4j version: {}. {}.'
                        'Skipping test.'.format(os.environ['NEO4J_VERSION'], message))


def basic_type_assertions(ground_truth, tested_object, test_description, check_neo4j_points=False):
    """
    Tests that `tested_object` has been created as intended.

    :param ground_truth: The object as it is supposed to have been created.
    :type ground_truth: NeomodelPoint or neo4j.v1.spatial.Point
    :param tested_object: The object as it results from one of the contructors.
    :type tested_object: NeomodelPoint or neo4j.v1.spatial.Point
    :param test_description: A brief description of the test being performed.
    :type test_description: str
    :param check_neo4j_points: Whether to assert between NeomodelPoint or neo4j.v1.spatial.Point objects.
    :type check_neo4j_points: bool
    :return:
    """
    if check_neo4j_points:
        assert isinstance(tested_object, type(ground_truth)), '{} did not return Neo4j Point'.format(test_description)
        assert tested_object.srid == ground_truth.srid, \
            '{} does not have the expected SRID({})'.format(test_description, ground_truth.srid)
        assert len(tested_object) == len(ground_truth), \
            '{} dimensionality mismatch. Expected {}, had {}'.format(len(ground_truth.coords),
                                                                     len(tested_object.coords))
    else:
        assert isinstance(tested_object, type(ground_truth)), '{} did not return NeomodelPoint'.format(test_description)
        assert tested_object.crs == ground_truth.crs, \
            '{} does not have the expected CRS({})'.format(test_description, ground_truth.crs)
        assert len(tested_object.coords[0]) == len(ground_truth.coords[0]), \
            '{} dimensionality mismatch. Expected {}, had {}'.format(len(ground_truth.coords[0]),
                                                                     len(tested_object.coords[0]))


# Object Construction
def test_coord_constructor():
    """
    Tests all the possible ways by which a NeomodelPoint can be instantiated successfully via passing coordinates.
    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(340, 'This version does not support spatial data types.')

    # Implicit cartesian point with coords
    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0))
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0))
    basic_type_assertions(ground_truth_object, new_point, 'Implicit 2d cartesian point instantiation')

    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0, 0.0))
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0, 0.0))
    basic_type_assertions(ground_truth_object, new_point, 'Implicit 3d cartesian point instantiation')

    # Explicit geographical point with coords
    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0), crs='wgs-84')
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0), crs='wgs-84')
    basic_type_assertions(ground_truth_object, new_point,
                          'Explicit 2d geographical point with tuple of coords instantiation')

    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0, 0.0), crs='wgs-84-3d')
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0, 0.0), crs='wgs-84-3d')
    basic_type_assertions(ground_truth_object, new_point,
                          'Explicit 3d geographical point with tuple of coords instantiation')

    # Cartesian point with named arguments
    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint(x=0.0, y=0.0)
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(x=0.0, y=0.0)
    basic_type_assertions(ground_truth_object, new_point, 'Cartesian 2d point with named arguments')

    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint(x=0.0, y=0.0, z=0.0)
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(x=0.0, y=0.0, z=0.0)
    basic_type_assertions(ground_truth_object, new_point, 'Cartesian 3d point with named arguments')

    # Geographical point with named arguments
    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint(longitude=0.0, latitude=0.0)
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(longitude=0.0, latitude=0.0)
    basic_type_assertions(ground_truth_object, new_point, 'Geographical 2d point with named arguments')

    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint(longitude=0.0, latitude=0.0, height=0.0)
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(longitude=0.0, latitude=0.0 ,height=0.0)
    basic_type_assertions(ground_truth_object, new_point, 'Geographical 3d point with named arguments')


def test_copy_constructors():
    """
    Tests all the possible ways by which a NeomodelPoint can be instantiated successfully via a copy constructor call.

    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(340, 'This version does not support spatial data types.')

    # Instantiate from Shapely point

    # Implicit cartesian from shapely point
    ground_truth = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0), crs='cartesian')
    shapely_point = shapely.geometry.Point((0.0, 0.0))
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(shapely_point)
    basic_type_assertions(ground_truth, new_point, 'Implicit cartesian by shapely Point')

    # Explicit geographical by shapely point
    ground_truth = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0, 0.0), crs='wgs-84-3d')
    shapely_point = shapely.geometry.Point((0.0, 0.0, 0.0))
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(shapely_point, crs='wgs-84-3d')
    basic_type_assertions(ground_truth, new_point, 'Explicit geographical by shapely Point')

    # Copy constructor for NeomodelPoints
    ground_truth = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0))
    other_neomodel_point = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0))
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(other_neomodel_point)
    basic_type_assertions(ground_truth, new_point, 'NeomodelPoint copy constructor')


def test_prohibited_constructor_forms():
    """
    Tests all the possible forms by which construction of NeomodelPoints should fail.

    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(340, 'This version does not support spatial data types.')

    # Absurd CRS
    with pytest.raises(ValueError, match=r'Invalid CRS\(blue_hotel\)'):
        new_point = neomodel.contrib.spatial_properties.NeomodelPoint((0,0), crs='blue_hotel')

    # Absurd coord dimensionality
    with pytest.raises(ValueError, match='Invalid vector dimensions. Expected 2 or 3, received 7'):
        new_point = neomodel.contrib.spatial_properties.NeomodelPoint((0,0,0,0,0,0,0), crs='cartesian')

    # Absurd datatype passed to copy constructor
    with pytest.raises(TypeError, match='Invalid object passed to copy constructor'):
        new_point = neomodel.contrib.spatial_properties.NeomodelPoint('it don''t mean a thing if it '
                                                                      'ain''t got that swing', crs='cartesian')

    # Trying to instantiate a point with any of BOTH x,y,z or longitude, latitude, height
    with pytest.raises(ValueError, match='Invalid instantiation via arguments'):
        new_point = neomodel.contrib.spatial_properties.NeomodelPoint(x=0.0, y=0.0,
                                                                      longitude=0.0, latitude=2.0, height=-2.0,
                                                                      crs='cartesian')

    # Trying to instantiate a point with absolutely NO parameters
    with pytest.raises(ValueError, match='Invalid instantiation via no arguments'):
        new_point = neomodel.contrib.spatial_properties.NeomodelPoint()


def test_property_accessors_depending_on_crs():
    """
    Tests that points are accessed via their respective accessors.

    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(340, 'This version does not support spatial data types.')

    # Geometrical points only have x,y,z coordinates
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0, 0.0), crs='cartesian-3d')
    with pytest.raises(AttributeError, match=r'Invalid coordinate \("longitude"\)'):
        new_point.longitude
    with pytest.raises(AttributeError, match=r'Invalid coordinate \("latitude"\)'):
        new_point.latitude
    with pytest.raises(AttributeError, match=r'Invalid coordinate \("height"\)'):
        new_point.height

    # Geographical points only have longitude, latitude, height coordinates
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0, 0.0), crs='wgs-84-3d')
    with pytest.raises(AttributeError, match=r'Invalid coordinate \("x"\)'):
        new_point.x
    with pytest.raises(AttributeError, match=r'Invalid coordinate \("y"\)'):
        new_point.y
    with pytest.raises(AttributeError, match=r'Invalid coordinate \("z"\)'):
        new_point.z


def test_property_accessors():
    """
    Tests that points are accessed via their respective accessors and that these accessors return the right values.

    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(340, 'This version does not support spatial data types.')

    # Geometrical points
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 1.0, 2.0), crs='cartesian-3d')
    assert new_point.x == 0.0, 'Expected x coordinate to be 0.0'
    assert new_point.y == 1.0, 'Expected y coordinate to be 1.0'
    assert new_point.z == 2.0, 'Expected z coordinate to be 2.0'

    # Geographical points
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 1.0, 2.0), crs='wgs-84-3d')
    assert new_point.longitude == 0.0, 'Expected longitude to be 0.0'
    assert new_point.latitude == 1.0, 'Expected latitude to be 1.0'
    assert new_point.height == 2.0, 'Expected height to be 2.0'



from neomodel import StructuredNode, StringProperty, db
from multiprocessing.pool import ThreadPool as Pool

class ThingyMaBob(StructuredNode):
    name = StringProperty(unique_index=True, required=True)

def thing_create(name):
    name = str(name)
    thing, = ThingyMaBob.get_or_create({'name': name})
    return thing.name, name


def test_concurrency():
    with Pool(5) as p:
        results = p.map(thing_create, range(50))
        for returned, sent in results:
            assert returned == sent


from neomodel.exceptions import (
    AttemptedCardinalityViolation, CardinalityViolation
)
from neomodel.relationship_manager import (
    RelationshipManager, ZeroOrMore
)  #noqa: F401


class ZeroOrOne(RelationshipManager):
    """ A relationship to zero or one node. """
    description = "zero or one relationship"

    def single(self):
        """
        Return the associated node.

        :return: node
        """
        nodes = super(ZeroOrOne, self).all()
        if len(nodes) == 1:
            return nodes[0]
        if len(nodes) > 1:
            raise CardinalityViolation(self, len(nodes))

    def all(self):
        node = self.single()
        return [node] if node else []

    def connect(self, node, properties=None):
        """
        Connect to a node.

        :param node:
        :type: StructuredNode
        :param properties: relationship properties
        :type: dict
        :return: True / rel instance
        """
        if len(self):
            raise AttemptedCardinalityViolation(
                    "Node already has {0} can't connect more".format(self))
        else:
            return super(ZeroOrOne, self).connect(node, properties)


class OneOrMore(RelationshipManager):
    """ A relationship to zero or more nodes. """
    description = "one or more relationships"

    def single(self):
        """
        Fetch one of the related nodes

        :return: Node
        """
        nodes = super(OneOrMore, self).all()
        if nodes:
            return nodes[0]
        raise CardinalityViolation(self, 'none')

    def all(self):
        """
        Returns all related nodes.

        :return: [node1, node2...]
        """
        nodes = super(OneOrMore, self).all()
        if nodes:
            return nodes
        raise CardinalityViolation(self, 'none')

    def disconnect(self, node):
        """
        Disconnect node
        :param node:
        :return:
        """
        if super(OneOrMore, self).__len__() < 2:
            raise AttemptedCardinalityViolation("One or more expected")
        return super(OneOrMore, self).disconnect(node)


class One(RelationshipManager):
    """
    A relationship to a single node
    """
    description = "one relationship"

    def single(self):
        """
        Return the associated node.

        :return: node
        """
        nodes = super(One, self).all()
        if nodes:
            if len(nodes) == 1:
                return nodes[0]
            else:
                raise CardinalityViolation(self, len(nodes))
        else:
            raise CardinalityViolation(self, 'none')

    def all(self):
        """
        Return single node in an array

        :return: [node]
        """
        return [self.single()]

    def disconnect(self, node):
        raise AttemptedCardinalityViolation(
            "Cardinality one, cannot disconnect use reconnect."
        )

    def disconnect_all(self):
        raise AttemptedCardinalityViolation(
            "Cardinality one, cannot disconnect_all use reconnect."
        )

    def connect(self, node, properties=None):
        """
        Connect a node

        :param node:
        :param properties: relationship properties
        :return: True / rel instance
        """
        if not hasattr(self.source, 'id'):
            raise ValueError("Node has not been saved cannot connect!")
        if len(self):
            raise AttemptedCardinalityViolation(
                "Node already has one relationship"
            )
        else:
            return super(One, self).connect(node, properties)



from neomodel import StructuredNode, StringProperty

HOOKS_CALLED = {}


class HookTest(StructuredNode):
    name = StringProperty()

    def post_create(self):
        HOOKS_CALLED['post_create'] = 1

    def pre_save(self):
        HOOKS_CALLED['pre_save'] = 1

    def post_save(self):
        HOOKS_CALLED['post_save'] = 1

    def pre_delete(self):
        HOOKS_CALLED['pre_delete'] = 1

    def post_delete(self):
        HOOKS_CALLED['post_delete'] = 1


def test_hooks():
    ht = HookTest(name='k').save()
    ht.delete()
    assert 'pre_save' in HOOKS_CALLED
    assert 'post_save' in HOOKS_CALLED
    assert 'post_create' in HOOKS_CALLED
    assert 'pre_delete' in HOOKS_CALLED
    assert 'post_delete' in HOOKS_CALLED





# pep8: noqa
import pkg_resources
from .core import *
from neomodel.exceptions import *
from .util import clear_neo4j_database, change_neo4j_password
from neomodel.match import EITHER, INCOMING, OUTGOING, NodeSet, Traversal
from neomodel.match_q import Q  # noqa
from neomodel.relationship_manager import (
    NotConnected, RelationshipTo, RelationshipFrom, Relationship,
    RelationshipManager, RelationshipDefinition
)
from .relationship import StructuredRel
from .cardinality import (ZeroOrMore, OneOrMore, ZeroOrOne, One)
from .properties import (StringProperty, IntegerProperty, AliasProperty,
                         FloatProperty, BooleanProperty, 
                         DateTimeFormatProperty, DateTimeProperty,
                         DateProperty, NormalizedProperty, RegexProperty,
                         EmailProperty, JSONProperty, ArrayProperty,
                         UniqueIdProperty)

__author__ = 'Robin Edwards'
__email__ = 'robin.ge@gmail.com'
__license__ = 'MIT'
__package__ = 'neomodel'
__version__ = '4.0.10'




from operator import itemgetter

from neomodel import StructuredNode, StringProperty, config
from neomodel.core import db, remove_all_labels
from neo4j.exceptions import ClientError

config.AUTO_INSTALL_LABELS = True


class ConstraintAndIndex(StructuredNode):
    name = StringProperty(unique_index=True)
    last_name = StringProperty(index=True)


def test_drop_labels():
    constraints_before, meta = db.cypher_query("CALL db.constraints()")
    indexes_before, meta = db.cypher_query("CALL db.indexes()")

    assert len(constraints_before) > 0
    assert len(indexes_before) > 0

    remove_all_labels()

    constraints, meta = db.cypher_query("CALL db.constraints()")
    indexes, meta = db.cypher_query("CALL db.indexes()")

    assert len(constraints) == 0
    # Ignore the automatically created LOOKUP indexes
    assert len([index for index in indexes if index[7] != []]) == 0

    # Returning all old constraints and indexes
    # Versions prior to 4.0 have a very different return format
    for constraint in constraints_before:
        if constraint[0].startswith('CONSTRAINT '):
            db.cypher_query('CREATE ' + constraint[0])
        else:
            db.cypher_query('CREATE ' + constraint[1])
    for index in indexes_before:
        try:
            if not isinstance(index[0], int) and index[0].startswith('INDEX '):
                db.cypher_query('CREATE ' + index[0])
            else:
                # Ignore the automatically created LOOKUP indexes
                if index[7] == []:
                    continue
                db.cypher_query('CREATE INDEX {0} FOR (n:{1}) ON (n.{2})'.format(index[1], index[7][0], index[8][0]))
        except ClientError:
            pass



from __future__ import print_function
import warnings
import os
import sys

import pytest

from neomodel import config, db, clear_neo4j_database, change_neo4j_password
from neo4j.exceptions import ClientError as CypherError
from neobolt.exceptions import ClientError


def pytest_addoption(parser):
    """
    Adds the command line option --resetdb.
    
    :param parser: The parser object. Please see <https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_addoption>`_
    :type Parser object: For more information please see <https://docs.pytest.org/en/latest/reference.html#_pytest.config.Parser>`_
    """
    parser.addoption("--resetdb", action="store_true", help = "Ensures that the database is clear prior to running tests for neomodel", default=False)
    

def pytest_sessionstart(session):
    """
    Provides initial connection to the database and sets up the rest of the test suite
    
    :param session: The session object. Please see <https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_sessionstart>`_
    :type Session object: For more information please see <https://docs.pytest.org/en/latest/reference.html#session>`_
    """
    
    warnings.simplefilter('default')
    
    config.DATABASE_URL = os.environ.get('NEO4J_BOLT_URL', 'bolt://neo4j:foobarbaz@localhost:7687')
    config.AUTO_INSTALL_LABELS = True
    
    try:
        # Clear the database if required
        database_is_populated, _ = db.cypher_query("MATCH (a) return count(a)>0 as database_is_populated")
        if database_is_populated[0][0] and not session.config.getoption("resetdb"):
            raise SystemError("Please note: The database seems to be populated.\n\tEither delete all nodes and edges manually, or set the --resetdb parameter when calling pytest\n\n\tpytest --resetdb.")
        else:
            clear_neo4j_database(db, clear_constraints=True, clear_indexes=True)        
    except (CypherError, ClientError) as ce:
        # Handle instance without password being changed
        if 'The credentials you provided were valid, but must be changed before you can use this instance' in str(ce):
            warnings.warn("New database with no password set, setting password to 'test'")
            try:
                change_neo4j_password(db, 'test')
                # Ensures that multiprocessing tests can use the new password
                config.DATABASE_URL = 'bolt://neo4j:test@localhost:7687'
                db.set_connection('bolt://neo4j:test@localhost:7687')
                warnings.warn("Please 'export NEO4J_BOLT_URL=bolt://neo4j:test@localhost:7687' for subsequent test runs")
            except (CypherError, ClientError) as e:
                if 'The credentials you provided were valid, but must be changed before you can use this instance' in str(e):
                    warnings.warn("You appear to be running on version 4.0+ of Neo4j, without having changed the password."
                        "Please manually log in, change your password, then update the config.DATABASE_URL call at line 32 in this file")
                else:
                    raise e
        else:
            raise ce


def version_to_dec(a_version_string):
    """
    Converts a version string to a number to allow for quick checks on the versions of specific components.

    :param a_version_string: The version string under test (e.g. '3.4.0')
    :type a_version_string: str
    :return: An integer representation of the string version, e.g. '3.4.0' --> 340
    """
    components = a_version_string.split('.')
    while len(components) < 3:
        components.append('0')
    num = 0
    for a_component in enumerate(components):
        num += (10 ** ((len(components) - 1) - a_component[0])) * int(a_component[1])
    return num


def check_and_skip_neo4j_least_version(required_least_neo4j_version, message):
    """
    Checks if the NEO4J_VERSION is at least `required_least_neo4j_version` and skips a test if not.

    WARNING: If the NEO4J_VERSION variable is not set, this function returns True, allowing the test to go ahead.

    :param required_least_neo4j_version: The least version to check. This must be the numberic representation of the
    version. That is: '3.4.0' would be passed as 340.
    :type required_least_neo4j_version: int
    :param message: An informative message as to why the calling test had to be skipped.
    :type message: str
    :return: A boolean value of True if the version reported is at least `required_least_neo4j_version`
    """
    if 'NEO4J_VERSION' in os.environ:
        if version_to_dec(os.environ['NEO4J_VERSION']) < required_least_neo4j_version:
            pytest.skip('Neo4j version: {}. {}.'
                        'Skipping test.'.format(os.environ['NEO4J_VERSION'], message))

@pytest.fixture
def skip_neo4j_before_330():
    check_and_skip_neo4j_least_version(330, 'Neo4J version does not support this test')



from pytest import raises

from neomodel import (StructuredNode, StringProperty, IntegerProperty, UniqueIdProperty,
                      RelationshipTo, RelationshipFrom, config)
from neomodel.exceptions import UniqueProperty, DeflateError

config.AUTO_INSTALL_LABELS = True

class UniqueUser(StructuredNode):
    uid = UniqueIdProperty()
    name = StringProperty()
    age = IntegerProperty()


def test_unique_id_property_batch():
    users = UniqueUser.create(
        {'name': 'bob', 'age': 2},
        {'name': 'ben', 'age': 3}
    )

    assert users[0].uid != users[1].uid

    users = UniqueUser.get_or_create(
        {'uid': users[0].uid},
        {'name': 'bill', 'age': 4}
    )

    assert users[0].name == 'bob'
    assert users[1].uid


class Customer(StructuredNode):
    email = StringProperty(unique_index=True, required=True)
    age = IntegerProperty(index=True)


def test_batch_create():
    users = Customer.create(
        {'email': 'jim1@aol.com', 'age': 11},
        {'email': 'jim2@aol.com', 'age': 7},
        {'email': 'jim3@aol.com', 'age': 9},
        {'email': 'jim4@aol.com', 'age': 7},
        {'email': 'jim5@aol.com', 'age': 99},
    )
    assert len(users) == 5
    assert users[0].age == 11
    assert users[1].age == 7
    assert users[1].email == 'jim2@aol.com'
    assert Customer.nodes.get(email='jim1@aol.com')


def test_batch_create_or_update():
    users = Customer.create_or_update(
        {'email': 'merge1@aol.com', 'age': 11},
        {'email': 'merge2@aol.com'},
        {'email': 'merge3@aol.com', 'age': 1},
        {'email': 'merge2@aol.com', 'age': 2},
    )
    assert len(users) == 4
    assert users[1] == users[3]
    assert Customer.nodes.get(email='merge1@aol.com').age == 11

    more_users = Customer.create_or_update(
        {'email': 'merge1@aol.com', 'age': 22},
        {'email': 'merge4@aol.com', 'age': None}
    )
    assert len(more_users) == 2
    assert users[0] == more_users[0]
    assert Customer.nodes.get(email='merge1@aol.com').age == 22


def test_batch_validation():
    # test validation in batch create
    with raises(DeflateError):
        Customer.create(
            {'email': 'jim1@aol.com', 'age': 'x'},
        )


def test_batch_index_violation():
    for u in Customer.nodes.all():
        u.delete()

    users = Customer.create(
        {'email': 'jim6@aol.com', 'age': 3},
    )
    assert users
    with raises(UniqueProperty):
        Customer.create(
            {'email': 'jim6@aol.com', 'age': 3},
            {'email': 'jim7@aol.com', 'age': 5},
        )

    # not found
    assert not Customer.nodes.filter(email='jim7@aol.com')


class Dog(StructuredNode):
    name = StringProperty(required=True)
    owner = RelationshipTo('Person', 'owner')


class Person(StructuredNode):
    name = StringProperty(unique_index=True)
    pets = RelationshipFrom('Dog', 'owner')


def test_get_or_create_with_rel():
    bob = Person.get_or_create({"name": "Bob"})[0]
    bobs_gizmo = Dog.get_or_create({"name": "Gizmo"}, relationship=bob.pets)

    tim = Person.get_or_create({"name": "Tim"})[0]
    tims_gizmo = Dog.get_or_create({"name": "Gizmo"}, relationship=tim.pets)

    # not the same gizmo
    assert bobs_gizmo[0] != tims_gizmo[0]



from datetime import datetime

from pytest import raises
import pytz

from neomodel import (StructuredNode, StructuredRel, Relationship, RelationshipTo,
                      StringProperty, DateTimeProperty, DeflateError)

HOOKS_CALLED = {
    'pre_save': 0,
    'post_save': 0
}


class FriendRel(StructuredRel):
    since = DateTimeProperty(default=lambda: datetime.now(pytz.utc))


class HatesRel(FriendRel):
    reason = StringProperty()

    def pre_save(self):
        HOOKS_CALLED['pre_save'] += 1

    def post_save(self):
        HOOKS_CALLED['post_save'] += 1


class Badger(StructuredNode):
    name = StringProperty(unique_index=True)
    friend = Relationship('Badger', 'FRIEND', model=FriendRel)
    hates = RelationshipTo('Stoat', 'HATES', model=HatesRel)


class Stoat(StructuredNode):
    name = StringProperty(unique_index=True)
    hates = RelationshipTo('Badger', 'HATES', model=HatesRel)


def test_either_connect_with_rel_model():
    paul = Badger(name="Paul").save()
    tom = Badger(name="Tom").save()

    # creating rels
    new_rel = tom.friend.disconnect(paul)
    new_rel = tom.friend.connect(paul)
    assert isinstance(new_rel, FriendRel)
    assert isinstance(new_rel.since, datetime)

    # updating properties
    new_rel.since = datetime.now(pytz.utc)
    assert isinstance(new_rel.save(), FriendRel)

    # start and end nodes are the opposite of what you'd expect when using either..
    # I've tried everything possible to correct this to no avail
    paul = new_rel.start_node()
    tom = new_rel.end_node()
    assert paul.name == 'Tom'
    assert tom.name == 'Paul'


def test_direction_connect_with_rel_model():
    paul = Badger(name="Paul the badger").save()
    ian = Stoat(name="Ian the stoat").save()

    rel = ian.hates.connect(paul, {'reason': "thinks paul should bath more often"})
    assert isinstance(rel.since, datetime)
    assert isinstance(rel, FriendRel)
    assert rel.reason.startswith("thinks")
    rel.reason = 'he smells'
    rel.save()

    ian = rel.start_node()
    assert isinstance(ian, Stoat)
    paul = rel.end_node()
    assert isinstance(paul, Badger)

    assert ian.name.startswith("Ian")
    assert paul.name.startswith("Paul")

    rel = ian.hates.relationship(paul)
    assert isinstance(rel, HatesRel)
    assert isinstance(rel.since, datetime)
    rel.save()

    # test deflate checking
    rel.since = "2:30pm"
    with raises(DeflateError):
        rel.save()

    # check deflate check via connect
    with raises(DeflateError):
        paul.hates.connect(ian, {'reason': "thinks paul should bath more often", 'since': '2:30pm'})


def test_traversal_where_clause():
    phill = Badger(name="Phill the badger").save()
    tim = Badger(name="Tim the badger").save()
    bob = Badger(name="Bob the badger").save()
    rel = tim.friend.connect(bob)
    now = datetime.now(pytz.utc)
    assert rel.since < now
    rel2 = tim.friend.connect(phill)
    assert rel2.since > now
    friends = tim.friend.match(since__gt=now)
    assert len(friends) == 1


def test_multiple_rels_exist_issue_223():
    # check a badger can dislike a stoat for multiple reasons
    phill = Badger(name="Phill").save()
    ian = Stoat(name="Stoat").save()

    rel_a = phill.hates.connect(ian, {'reason': 'a'})
    rel_b = phill.hates.connect(ian, {'reason': 'b'})
    assert rel_a.id != rel_b.id

    ian_a = phill.hates.match(reason='a')[0]
    ian_b = phill.hates.match(reason='b')[0]
    assert ian_a.id == ian_b.id


def test_retrieve_all_rels():
    tom = Badger(name="tom").save()
    ian = Stoat(name="ian").save()

    rel_a = tom.hates.connect(ian, {'reason': 'a'})
    rel_b = tom.hates.connect(ian, {'reason': 'b'})

    rels = tom.hates.all_relationships(ian)
    assert len(rels) == 2
    assert rels[0].id in [rel_a.id, rel_b.id]
    assert rels[1].id in [rel_a.id, rel_b.id]


def test_save_hook_on_rel_model():
    HOOKS_CALLED['pre_save'] = 0
    HOOKS_CALLED['post_save'] = 0

    paul = Badger(name="PaulB").save()
    ian = Stoat(name="IanS").save()

    rel = ian.hates.connect(paul, {'reason': "yadda yadda"})
    rel.save()

    assert HOOKS_CALLED['pre_save'] == 2
    assert HOOKS_CALLED['post_save'] == 2




from .properties import Property, PropertyManager
from .core import db
from .util import deprecated
from .hooks import hooks


class RelationshipMeta(type):
    def __new__(mcs, name, bases, dct):
        inst = super(RelationshipMeta, mcs).__new__(mcs, name, bases, dct)
        for key, value in dct.items():
            if issubclass(value.__class__, Property):
                value.name = key
                value.owner = inst
                if value.is_indexed:
                    raise NotImplemented("Indexed relationship properties not supported yet")

                # support for 'magic' properties
                if hasattr(value, 'setup') and hasattr(value.setup, '__call__'):
                    value.setup()
        return inst


StructuredRelBase = RelationshipMeta('RelationshipBase', (PropertyManager,), {})


class StructuredRel(StructuredRelBase):
    """
    Base class for relationship objects
    """
    def __init__(self, *args, **kwargs):
        super(StructuredRel, self).__init__(*args, **kwargs)

    @hooks
    def save(self):
        """
        Save the relationship

        :return: self
        """
        props = self.deflate(self.__properties__)
        query = "MATCH ()-[r]->() WHERE id(r)=$self "
        for key in props:
            query += " SET r.{0} = ${1}".format(key, key)
        props['self'] = self.id

        db.cypher_query(query, props)

        return self

    @deprecated('This method will be removed in neomodel 4')
    def delete(self):
        raise NotImplemented("Can not delete relationships please use"
                             " 'disconnect'")

    def start_node(self):
        """
        Get start node

        :return: StructuredNode
        """
        return db.cypher_query("MATCH (aNode) "
                               "WHERE id(aNode)={nodeid} "
                               "RETURN aNode".format(nodeid=self._start_node_id),
                               resolve_objects = True)[0][0][0]
      
    def end_node(self):
        """
        Get end node

        :return: StructuredNode
        """
        return db.cypher_query("MATCH (aNode) "
                               "WHERE id(aNode)={nodeid} "
                               "RETURN aNode".format(nodeid=self._end_node_id),
                               resolve_objects = True)[0][0][0]

    @classmethod
    def inflate(cls, rel):
        """
        Inflate a neo4j_driver relationship object to a neomodel object
        :param rel:
        :return: StructuredRel
        """
        props = {}
        for key, prop in cls.defined_properties(aliases=False, rels=False).items():
            if key in rel:
                props[key] = prop.inflate(rel[key], obj=rel)
            elif prop.has_default:
                props[key] = prop.default_value()
            else:
                props[key] = None
        srel = cls(**props)
        srel._start_node_id = rel.start_node.id
        srel._end_node_id = rel.end_node.id
        srel.id = rel.id
        return srel



from neomodel.core import StructuredNode
from neomodel.exceptions import InflateConflict, DeflateConflict
from neomodel.util import _get_node_properties


class SemiStructuredNode(StructuredNode):
    """
    A base class allowing properties to be stored on a node that aren't
    specified in its definition. Conflicting properties are signaled with the
    :class:`DeflateConflict` exception::

        class Person(SemiStructuredNode):
            name = StringProperty()
            age = IntegerProperty()

            def hello(self):
                print("Hi my names " + self.name)

        tim = Person(name='Tim', age=8, weight=11).save()
        tim.hello = "Hi"
        tim.save() # DeflateConflict
    """
    __abstract_node__ = True

    def __init__(self, *args, **kwargs):
        super(SemiStructuredNode, self).__init__(*args, **kwargs)

    @classmethod
    def inflate(cls, node):
        # support lazy loading
        if isinstance(node, int):
            snode = cls()
            snode.id = node
        else:
            props = {}
            node_properties = {}
            for key, prop in cls.__all_properties__:
                node_properties = _get_node_properties(node)
                if key in node_properties:
                    props[key] = prop.inflate(node_properties[key], node)
                elif prop.has_default:
                    props[key] = prop.default_value()
                else:
                    props[key] = None
            # handle properties not defined on the class
            for free_key in (x for x in node_properties if x not in props):
                if hasattr(cls, free_key):
                    raise InflateConflict(cls, free_key,
                                          node_properties[free_key], node.id)
                props[free_key] = node_properties[free_key]

            snode = cls(**props)
            snode.id = node.id

        return snode

    @classmethod
    def deflate(cls, node_props, obj=None, skip_empty=False):
        deflated = super(SemiStructuredNode, cls).deflate(node_props, obj,
                                                          skip_empty=skip_empty)
        for key in [k for k in node_props if k not in deflated]:
            if hasattr(cls, key):
                raise DeflateConflict(cls, key, deflated[key], obj.id)
        node_props.update(deflated)
        return node_props



import functools
import json
import sys
import types
import re
import uuid
import warnings
from datetime import date, datetime

import pytz

import neo4j.time

from neomodel import config
from neomodel.exceptions import InflateError, DeflateError, RequiredProperty


if sys.version_info >= (3, 0):
    unicode = str


def display_for(key):
    def display_choice(self):
        return getattr(self.__class__, key).choices[getattr(self, key)]
    return display_choice


class PropertyManager(object):
    """
    Common methods for handling properties on node and relationship objects.
    """

    def __init__(self, **kwargs):
        properties = getattr(self, "__all_properties__", None)
        if properties is None:
            properties = \
                self.defined_properties(rels=False, aliases=False).items()
        for name, property in properties:
            if kwargs.get(name) is None:
                if getattr(property, 'has_default', False):
                    setattr(self, name, property.default_value())
                else:
                    setattr(self, name, None)
            else:
                setattr(self, name, kwargs[name])

            if getattr(property, 'choices', None):
                setattr(self, 'get_{0}_display'.format(name),
                        types.MethodType(display_for(name), self))

            if name in kwargs:
                del kwargs[name]

        aliases = getattr(self, "__all_aliases__", None)
        if aliases is None:
            aliases = self.defined_properties(
                aliases=True, rels=False, properties=False).items()
        for name, property in aliases:
            if name in kwargs:
                setattr(self, name, kwargs[name])
                del kwargs[name]

        # undefined properties (for magic @prop.setters etc)
        for name, property in kwargs.items():
            setattr(self, name, property)

    @property
    def __properties__(self):
        from .relationship_manager import RelationshipManager

        return dict((name, value) for name, value in vars(self).items()
                    if not name.startswith('_')
                    and not callable(value)
                    and not isinstance(value,
                                       (RelationshipManager, AliasProperty,))
                    )

    @classmethod
    def deflate(cls, properties, obj=None, skip_empty=False):
        # deflate dict ready to be stored
        deflated = {}
        for name, property \
                in cls.defined_properties(aliases=False, rels=False).items():
            db_property = property.db_property or name
            if properties.get(name) is not None:
                deflated[db_property] = property.deflate(properties[name], obj)
            elif property.has_default:
                deflated[db_property] = property.deflate(
                    property.default_value(), obj
                )
            elif property.required:
                raise RequiredProperty(name, cls)
            elif not skip_empty:
                deflated[db_property] = None
        return deflated

    @classmethod
    def defined_properties(cls, aliases=True, properties=True, rels=True):
        from .relationship_manager import RelationshipDefinition
        props = {}
        for baseclass in reversed(cls.__mro__):
            props.update(dict(
                (name, property) for name, property in vars(baseclass).items()
                if (aliases and isinstance(property, AliasProperty))
                or (properties and isinstance(property, Property)
                    and not isinstance(property, AliasProperty))
                or (rels and isinstance(property, RelationshipDefinition))
            ))
        return props


def validator(fn):
    fn_name = fn.func_name if hasattr(fn, 'func_name') else fn.__name__
    if fn_name == 'inflate':
        exc_class = InflateError
    elif fn_name == 'deflate':
        exc_class = DeflateError
    else:
        raise Exception("Unknown Property method " + fn_name)

    @functools.wraps(fn)
    def _validator(self, value, obj=None, rethrow=True):
        if rethrow:
            try:
                return fn(self, value)
            except Exception as e:
                raise exc_class(self.name, self.owner, str(e), obj)
        else:
            # For using with ArrayProperty where we don't want an Inflate/Deflate error.
            return fn(self, value)

    return _validator


class Property(object):
    """
    Base class for object properties.

    :param unique_index: Creates a unique index for this property. Defaults to
                         ``False``.
    :type unique_index: :class:`bool`
    :param index: Creates an index for this property. Defaults to ``False``.
    :type index: :class:`bool`
    :param required: Marks the property as required. Defaults to ``False``.
    :type required: :class:`bool`
    :param default: A default value or callable that returns one to set when a
                    node is initialized without specifying this property.
    :param db_property: A name that this property maps to in the database.
                        Defaults to the model's property name.
    :type db_property: :class:`str`
    :param label: Optional, used by ``django_neomodel``.
    :type label: :class:`str`
    :param help_text: Optional, used by ``django_neomodel``.
    :type help_text: :class:`str`
    """

    form_field_class = 'CharField'

    def __init__(self, unique_index=False, index=False, required=False, default=None,
                 db_property=None, label=None, help_text=None, **kwargs):

        if default is not None and required:
            raise ValueError(
                "The arguments `required` and `default` are mutually exclusive."
            )
        if unique_index and index:
            raise ValueError(
                "The arguments `unique_index` and `index` are mutually exclusive."
            )

        self.required = required
        self.unique_index = unique_index
        self.index = index
        self.default = default
        self.has_default = True if self.default is not None else False
        self.db_property = db_property
        self.label = label
        self.help_text = help_text

    def default_value(self):
        """
        Generate a default value

        :return: the value
        """
        if self.has_default:
            if hasattr(self.default, '__call__'):
                return self.default()
            else:
                return self.default
        else:
            raise Exception("No default value specified")

    @property
    def is_indexed(self):
        return self.unique_index or self.index


class NormalizedProperty(Property):
    """
    Base class for normalized properties. These use the same normalization
    method to in- or deflating.
    """

    @validator
    def inflate(self, value):
        return self.normalize(value)

    @validator
    def deflate(self, value):
        return self.normalize(value)

    def default_value(self):
        default = super(NormalizedProperty, self).default_value()
        return self.normalize(default)

    def normalize(self, value):
        raise NotImplementedError('Specialize normalize method')


# TODO remove this with the next major release
def _warn_NormalProperty_renamed():
    warnings.warn(
        'The class NormalProperty was renamed to NormalizedProperty. '
        'Use that one as base class. The former will be removed in the next '
        'major release.', DeprecationWarning)


if sys.version_info >= (3, 6):
    class NormalProperty(NormalizedProperty):

        def __init_subclass__(cls, **kwargs):
            _warn_NormalProperty_renamed()
else:
    class NormalProperty(NormalizedProperty):

        def __init__(self, *args, **kwargs):
            _warn_NormalProperty_renamed()
            super(NormalProperty, self).__init__(*args, **kwargs)
##


class RegexProperty(NormalizedProperty):
    """
    Validates a property against a regular expression.

    If sub-classing set:

        expression = r'[^@]+@[^@]+\.[^@]+'
    """
    form_field_class = 'RegexField'

    expression = None

    def __init__(self, expression=None, **kwargs):
        """
        Initializes new property with an expression.

        :param str expression: regular expression validating this property
        """
        super(RegexProperty, self).__init__(**kwargs)
        actual_re = expression or self.expression
        if actual_re is None:
            raise ValueError('expression is undefined')
        self.expression = actual_re

    def normalize(self, value):
        normal = unicode(value)
        if not re.match(self.expression, normal):
            raise ValueError(
                '{0!r} does not matches {1!r}'.format(
                    value,
                    self.expression,
                )
            )
        return normal


class EmailProperty(RegexProperty):
    """
    Store email addresses
    """
    form_field_class = 'EmailField'
    expression = r'[^@]+@[^@]+\.[^@]+'


class StringProperty(NormalizedProperty):
    """
    Stores a unicode string

    :param choices: A mapping of valid strings to label strings that are used
                    to display information in an application. If the default
                    value ``None`` is used, any string is valid.
    :type choices: Any type that can be used to initiate a :class:`dict`.
    :param max_length: The maximum non-zero length that this attribute can be
    :type max_length: int
    """

    def __init__(self, choices=None, max_length=None, **kwargs):
        if max_length is not None:
            if choices is not None:
                raise ValueError("The arguments `choices` and `max_length` are mutually exclusive.")
            if max_length<1:
                raise ValueError("`max_length` cannot be zero or take negative values.")

        super(StringProperty, self).__init__(**kwargs)

        self.max_length = max_length
        if choices is None:
            self.choices = None
        else:
            try:
                self.choices = dict(choices)
            except Exception:
                raise ValueError("The choices argument must be convertable to a dictionary.")
            # Python 3:
            except Exception as e:
                raise ValueError("The choices argument must be convertable to "
                                 "a dictionary.") from e
            self.form_field_class = 'TypedChoiceField'

    def normalize(self, value):
        # One thing to note here is that the following two checks can remain uncoupled
        # as long as it is guaranteed (by the constructor) that `choices` and `max_length`
        # are mutually exclusive. If that check in the constructor ever has to be removed, 
        # these two validation checks here will have to be coupled so that having set 
        # `choices` overrides having set the `max_length`.
        if self.choices is not None and value not in self.choices:
            raise ValueError("Invalid choice: {}".format(value))
        if self.max_length is not None and len(value) > self.max_length:
            raise ValueError("Property max length exceeded. Expected {}, got {} == len('{}')".format(
                             self.max_length, len(value), value))
        return unicode(value)

    def default_value(self):
        return self.normalize(super(StringProperty, self).default_value())


class IntegerProperty(Property):
    """
    Stores an Integer value
    """
    form_field_class = 'IntegerField'

    @validator
    def inflate(self, value):
        return int(value)

    @validator
    def deflate(self, value):
        return int(value)

    def default_value(self):
        return int(super(IntegerProperty, self).default_value())


class ArrayProperty(Property):
    """
    Stores a list of items
    """

    def __init__(self, base_property=None, **kwargs):
        """
        Store a list of values, optionally of a specific type.

        :param base_property: List item type e.g StringProperty for string
        :type: Property
        """

        # list item type
        if base_property is not None:
            if not isinstance(base_property, Property):
                raise TypeError('Expecting neomodel Property')

            if isinstance(base_property, ArrayProperty):
                raise TypeError('Cannot have nested ArrayProperty')

            for ilegal_attr in ['default', 'index', 'unique_index', 'required']:
                if getattr(base_property, ilegal_attr, None):
                    raise ValueError('ArrayProperty base_property cannot have "{0}" set'.format(ilegal_attr))

        self.base_property = base_property

        super(ArrayProperty, self).__init__(**kwargs)

    @validator
    def inflate(self, value):
        if self.base_property:
            return [self.base_property.inflate(item, rethrow=False) for item in value]

        return list(value)

    @validator
    def deflate(self, value):
        if self.base_property:
            return [self.base_property.deflate(item, rethrow=False) for item in value]

        return list(value)

    def default_value(self):
        return list(super(ArrayProperty, self).default_value())


class FloatProperty(Property):
    """
    Store a floating point value
    """
    form_field_class = 'FloatField'

    @validator
    def inflate(self, value):
        return float(value)

    @validator
    def deflate(self, value):
        return float(value)

    def default_value(self):
        return float(super(FloatProperty, self).default_value())


class BooleanProperty(Property):
    """
    Stores a boolean value
    """
    form_field_class = 'BooleanField'

    @validator
    def inflate(self, value):
        return bool(value)

    @validator
    def deflate(self, value):
        return bool(value)

    def default_value(self):
        return bool(super(BooleanProperty, self).default_value())


class DateProperty(Property):
    """
    Stores a date
    """
    form_field_class = 'DateField'

    @validator
    def inflate(self, value):
        if isinstance(value, neo4j.time.DateTime):
            value = date(value.year, value.month, value.day)
        elif isinstance(value, str):
            if "T" in value:
                value = value[:value.find('T')]
        return datetime.strptime(unicode(value), "%Y-%m-%d").date()

    @validator
    def deflate(self, value):
        if not isinstance(value, date):
            msg = 'datetime.date object expected, got {0}'.format(repr(value))
            raise ValueError(msg)
        return value.isoformat()

class DateTimeFormatProperty(Property):
    """
    Store a datetime by custome format
    :param default_now: If ``True``, the creation time (Local) will be used as default.
                        Defaults to ``False``.
    :param format:      Date format string, default is %Y-%m-%d

    :type default_now:  :class:`bool`
    :type format:       :class:`str`
    """
    form_field_class = 'DateTimeFormatField'

    def __init__(self, default_now=False, format="%Y-%m-%d", **kwargs):
        if default_now:
            if 'default' in kwargs:
                raise ValueError('too many defaults')
            kwargs['default'] = lambda: datetime.now()

        self.format = format
        super(DateTimeFormatProperty, self).__init__(**kwargs)

    @validator
    def inflate(self, value):
        return datetime.strptime(unicode(value), self.format)

    @validator
    def deflate(self, value):
        if not isinstance(value, datetime):
            raise ValueError('datetime object expected, got {0}.'.format(type(value)))
        return datetime.strftime(value, self.format)




class DateTimeProperty(Property):
    """ A property representing a :class:`datetime.datetime` object as
        unix epoch.

        :param default_now: If ``True``, the creation time (UTC) will be used as default.
                            Defaults to ``False``.
        :type default_now: :class:`bool`
    """
    form_field_class = 'DateTimeField'

    def __init__(self, default_now=False, **kwargs):
        if default_now:
            if 'default' in kwargs:
                raise ValueError('too many defaults')
            kwargs['default'] = lambda: datetime.utcnow().replace(tzinfo=pytz.utc)

        super(DateTimeProperty, self).__init__(**kwargs)

    @validator
    def inflate(self, value):
        try:
            epoch = float(value)
        except ValueError:
            raise ValueError("Float or integer expected, got {0} can't inflate to "
                             "datetime.".format(type(value)))
        except TypeError:
            raise TypeError(
                "Float or integer expected. Can't inflate {0} to datetime.".format(type(value)))
        return datetime.utcfromtimestamp(epoch).replace(tzinfo=pytz.utc)

    @validator
    def deflate(self, value):
        if not isinstance(value, datetime):
            raise ValueError('datetime object expected, got {0}.'.format(type(value)))
        if value.tzinfo:
            value = value.astimezone(pytz.utc)
            epoch_date = datetime(1970, 1, 1, tzinfo=pytz.utc)
        elif config.FORCE_TIMEZONE:
            raise ValueError("Error deflating {0}: No timezone provided.".format(value))
        else:
            # No timezone specified on datetime object.. assuming UTC
            epoch_date = datetime(1970, 1, 1)
        return float((value - epoch_date).total_seconds())


class JSONProperty(Property):
    """
    Store a data structure as a JSON string.

    The structure will be inflated when a node is retrieved.
    """

    def __init__(self, *args, **kwargs):
        super(JSONProperty, self).__init__(*args, **kwargs)

    @validator
    def inflate(self, value):
        return json.loads(value)

    @validator
    def deflate(self, value):
        return json.dumps(value)


class AliasProperty(property, Property):
    """
    Alias another existing property
    """

    def __init__(self, to=None):
        """
        Create new alias

        :param to: name of property aliasing
        :type: str
        """
        self.target = to
        self.required = False
        self.has_default = False

    def aliased_to(self):
        return self.target

    def __get__(self, obj, cls):
        return getattr(obj, self.aliased_to()) if obj else self

    def __set__(self, obj, value):
        setattr(obj, self.aliased_to(), value)

    @property
    def index(self):
        return getattr(self.owner, self.aliased_to()).index

    @property
    def unique_index(self):
        return getattr(self.owner, self.aliased_to()).unique_index


class UniqueIdProperty(Property):
    """
    A unique identifier, a randomly generated uid (uuid4) with a unique index
    """

    def __init__(self, **kwargs):
        for item in ['required', 'unique_index', 'index', 'default']:
            if item in kwargs:
                raise ValueError('{0} argument ignored by {1}'.format(item, self.__class__.__name__))

        kwargs['unique_index'] = True
        kwargs['default'] = lambda: uuid.uuid4().hex
        super(UniqueIdProperty, self).__init__(**kwargs)

    @validator
    def inflate(self, value):
        return unicode(value)

    @validator
    def deflate(self, value):
        return unicode(value)



from neomodel import StructuredNode, RelationshipTo


class SomeModel(StructuredNode):
    test = RelationshipTo('SomeModel', 'SELF')


def test_len_relationship():
    t1 = SomeModel().save()
    t2 = SomeModel().save()

    t1.test.connect(t2)
    l = len(t1.test.all())

    assert l
    assert l == 1



from __future__ import print_function

from datetime import datetime
from pytest import raises

from neomodel import DateProperty, StructuredNode, StringProperty, IntegerProperty
from neomodel.exceptions import RequiredProperty, UniqueProperty


class User(StructuredNode):
    email = StringProperty(unique_index=True, required=True)
    age = IntegerProperty(index=True)

    @property
    def email_alias(self):
        return self.email

    @email_alias.setter # noqa
    def email_alias(self, value):
        self.email = value


def test_issue_233():
    class BaseIssue233(StructuredNode):
        __abstract_node__ = True

        def __getitem__(self, item):
            return self.__dict__[item]

    class Issue233(BaseIssue233):
        uid = StringProperty(unique_index=True, required=True)

    i = Issue233(uid='testgetitem').save()
    assert i['uid'] == 'testgetitem'


def test_issue_72():
    user = User(email='foo@bar.com')
    assert user.age is None


def test_required():
    try:
        User(age=3).save()
    except RequiredProperty:
        assert True
    else:
        assert False


def test_repr_and_str():
    u = User(email='robin@test.com', age=3)
    print(repr(u))
    print(str(u))
    assert True


def test_get_and_get_or_none():
    u = User(email='robin@test.com', age=3)
    assert u.save()
    rob = User.nodes.get(email='robin@test.com')
    assert rob.email == 'robin@test.com'
    assert rob.age == 3

    rob = User.nodes.get_or_none(email='robin@test.com')
    assert rob.email == 'robin@test.com'

    n = User.nodes.get_or_none(email='robin@nothere.com')
    assert n is None


def test_first_and_first_or_none():
    u = User(email='matt@test.com', age=24)
    assert u.save()
    u2 = User(email='tbrady@test.com', age=40)
    assert u2.save()
    tbrady = User.nodes.order_by('-age').first()
    assert tbrady.email == 'tbrady@test.com'
    assert tbrady.age == 40

    tbrady = User.nodes.order_by('-age').first_or_none()
    assert tbrady.email == 'tbrady@test.com'

    n = User.nodes.first_or_none(email='matt@nothere.com')
    assert n is None


def test_save_to_model():
    u = User(email='jim@test.com', age=3)
    assert u.save()
    assert u.id > 0
    assert u.email == 'jim@test.com'
    assert u.age == 3


def test_unique():
    User(email='jim1@test.com', age=3).save()
    try:
        User(email='jim1@test.com', age=3).save()
    except Exception as e:
        assert e.__class__.__name__ == 'UniqueProperty'
    else:
        assert False, "No exception raised."


def test_update_unique():
    u = User(email='jimxx@test.com', age=3).save()
    u.save()  # this shouldn't fail


def test_update():
    user = User(email='jim2@test.com', age=3).save()
    assert user
    user.email = 'jim2000@test.com'
    user.save()
    jim = User.nodes.get(email='jim2000@test.com')
    assert jim
    assert jim.email == 'jim2000@test.com'


def test_save_through_magic_property():
    user = User(email_alias='blah@test.com', age=8).save()
    assert user.email_alias == 'blah@test.com'
    user = User.nodes.get(email='blah@test.com')
    assert user.email == 'blah@test.com'
    assert user.email_alias == 'blah@test.com'

    user1 = User(email='blah1@test.com', age=8).save()
    assert user1.email_alias == 'blah1@test.com'
    user1.email_alias = 'blah2@test.com'
    assert user1.save()
    user2 = User.nodes.get(email='blah2@test.com')
    assert user2


class Customer2(StructuredNode):
    __label__ = 'customers'
    email = StringProperty(unique_index=True, required=True)
    age = IntegerProperty(index=True)


def test_not_updated_on_unique_error():
    Customer2(email='jim@bob.com', age=7).save()
    test = Customer2(email='jim1@bob.com', age=2).save()
    test.email = 'jim@bob.com'
    with raises(UniqueProperty):
        test.save()
    customers = Customer2.nodes.all()
    assert customers[0].email != customers[1].email
    assert Customer2.nodes.get(email='jim@bob.com').age == 7
    assert Customer2.nodes.get(email='jim1@bob.com').age == 2


def test_label_not_inherited():
    class Customer3(Customer2):
        address = StringProperty()

    assert Customer3.__label__ == 'Customer3'
    c = Customer3(email='test@test.com').save()
    assert 'customers' in c.labels()
    assert 'Customer3' in c.labels()

    c = Customer2.nodes.get(email='test@test.com')
    assert isinstance(c, Customer2)
    assert 'customers' in c.labels()
    assert 'Customer3' in c.labels()


def test_refresh():
    c = Customer2(email='my@email.com', age=16).save()
    c.my_custom_prop = 'value'
    copy = Customer2.nodes.get(email='my@email.com')
    copy.age = 20
    copy.save()

    assert c.age == 16

    c.refresh()
    assert c.age == 20
    assert c.my_custom_prop == 'value'

    c = Customer2.inflate(c.id)
    c.age = 30
    c.refresh()

    assert c.age == 20

    c = Customer2.inflate(999)
    with raises(Customer2.DoesNotExist):
        c.refresh()


def test_setting_value_to_none():
    c = Customer2(email='alice@bob.com', age=42).save()
    assert c.age is not None

    c.age = None
    c.save()

    copy = Customer2.nodes.get(email='alice@bob.com')
    assert copy.age is None


def test_inheritance():
    class User(StructuredNode):
        __abstract_node__ = True
        name = StringProperty(unique_index=True)

    class Shopper(User):
        balance = IntegerProperty(index=True)

        def credit_account(self, amount):
            self.balance = self.balance + int(amount)
            self.save()

    jim = Shopper(name='jimmy', balance=300).save()
    jim.credit_account(50)

    assert Shopper.__label__ == 'Shopper'
    assert jim.balance == 350
    assert len(jim.inherited_labels()) == 1
    assert len(jim.labels()) == 1
    assert jim.labels()[0] == 'Shopper'


def test_mixins():
    class UserMixin(object):
        name = StringProperty(unique_index=True)
        password = StringProperty()

    class CreditMixin(object):
        balance = IntegerProperty(index=True)

        def credit_account(self, amount):
            self.balance = self.balance + int(amount)
            self.save()

    class Shopper2(StructuredNode, UserMixin, CreditMixin):
        pass

    jim = Shopper2(name='jimmy', balance=300).save()
    jim.credit_account(50)

    assert Shopper2.__label__ == 'Shopper2'
    assert jim.balance == 350
    assert len(jim.inherited_labels()) == 1
    assert len(jim.labels()) == 1
    assert jim.labels()[0] == 'Shopper2'

def test_date_property():
    class DateTest(StructuredNode):
        birthdate = DateProperty()
    
    user = DateTest(birthdate=datetime.now()).save()




import pytest
from neobolt.addressing import AddressError
from pytest import raises

from neomodel import db, StructuredNode, StringProperty, UniqueProperty


class APerson(StructuredNode):
    name = StringProperty(unique_index=True)


def test_rollback_and_commit_transaction():
    for p in APerson.nodes:
        p.delete()

    APerson(name='Roger').save()

    db.begin()
    APerson(name='Terry S').save()
    db.rollback()

    assert len(APerson.nodes) == 1

    db.begin()
    APerson(name='Terry S').save()
    db.commit()

    assert len(APerson.nodes) == 2


@db.transaction
def in_a_tx(*names):
    for n in names:
        APerson(name=n).save()


def test_transaction_decorator():
    for p in APerson.nodes:
        p.delete()

    # should work
    in_a_tx('Roger')
    assert True

    # should bail but raise correct error
    with raises(UniqueProperty):
        in_a_tx('Jim', 'Roger')

    assert 'Jim' not in [p.name for p in APerson.nodes]


def test_transaction_as_a_context():
    with db.transaction:
        APerson(name='Tim').save()

    assert APerson.nodes.filter(name='Tim')

    with raises(UniqueProperty):
        with db.transaction:
            APerson(name='Tim').save()


def test_query_inside_transaction():
    for p in APerson.nodes:
        p.delete()

    with db.transaction:
        APerson(name='Alice').save()
        APerson(name='Bob').save()

        assert len([p.name for p in APerson.nodes]) == 2


def test_set_connection_works():
    assert APerson(name='New guy 1').save()
    from socket import gaierror

    old_url = db.url
    with raises(ValueError):
        db.set_connection('bolt://user:password@6.6.6.6.6.6.6.6:7687')
        APerson(name='New guy 2').save()
    db.set_connection(old_url)
    # set connection back
    assert APerson(name='New guy 3').save()

@db.transaction.with_bookmark
def in_a_tx(*names):
    for n in names:
        APerson(name=n).save()


def test_bookmark_transaction_decorator(skip_neo4j_before_330):
    for p in APerson.nodes:
        p.delete()

    # should work
    result, bookmark = in_a_tx('Ruth', bookmarks=None)
    assert result is None
    assert isinstance(bookmark, str)

    # should bail but raise correct error
    with raises(UniqueProperty):
        in_a_tx('Jane', 'Ruth')

    assert 'Jane' not in [p.name for p in APerson.nodes]


def test_bookmark_transaction_as_a_context(skip_neo4j_before_330):
    with db.transaction as transaction:
        APerson(name='Tanya').save()
    assert isinstance(transaction.last_bookmark, str)

    assert APerson.nodes.filter(name='Tanya')

    with raises(UniqueProperty):
        with db.transaction as transaction:
            APerson(name='Tanya').save()
    assert not hasattr(transaction, 'last_bookmark')

@pytest.fixture
def spy_on_db_begin(monkeypatch):
    spy_calls = []
    original_begin = db.begin

    def begin_spy(*args, **kwargs):
        spy_calls.append((args, kwargs))
        return original_begin(*args, **kwargs)

    monkeypatch.setattr(db, 'begin', begin_spy)
    return spy_calls

def test_bookmark_passed_in_to_context(skip_neo4j_before_330, spy_on_db_begin):
    transaction = db.transaction
    with transaction:
        pass

    assert spy_on_db_begin[-1] == ((), { 'access_mode': None })
    last_bookmark = transaction.last_bookmark

    transaction.bookmarks = last_bookmark
    with transaction:
        pass
    assert spy_on_db_begin[-1] == ((), { 'access_mode': None, 'bookmarks': (last_bookmark,) })

    transaction.bookmarks = [last_bookmark]
    with transaction:
        pass
    assert spy_on_db_begin[-1] == ((), { 'access_mode': None, 'bookmarks': (last_bookmark,) })

def test_query_inside_bookmark_transaction(skip_neo4j_before_330):
    for p in APerson.nodes:
        p.delete()

    with db.transaction as transaction:
        APerson(name='Alice').save()
        APerson(name='Bob').save()

        assert len([p.name for p in APerson.nodes]) == 2

    assert isinstance(transaction.last_bookmark, str)
